"""Definition of AsymmBand class. 

"""

import numpy as np
from fcband import gbutil
import sys
from fcband.gbconst import *
import re
from numpy import exp, pi, cos   
from scipy.integrate import quad 
from scipy import Inf            
from scipy.optimize import minimize


class AsymmBand(object):
    """Class for computing overall vibronic shape of a single band.

    """
    def __init__(self): 
        self.natom = None 
        self.ncart = None 
        self.atomnum = None 
        self.misc = {'geomformat':'g09',
                     'freqformat':'g09','gradformat':'g09'}
        #-- GS equilibrium geom.
        self.x0 = None 
        #-- ES equilibrium geom.
        self.xe = None 
        self.q0 = None 
        self.grad = None 
        self.freq = None 
        self.eigvec = None 
        #-- energy in eV
        self.Evert = None 
        self.deltaE = None 
        self.bgeomes = False 
        self.oscstr = None
        self.tdm2 = None

        self.m = None 
        self.bigq = None
        self.al = None
        self.ga = None

        self.start = None 
        self.end = None 
        self.npoints = None 
        self.unit = 'ev'


    def readstuff(self, fin, keyw):
        """Read in a piece of data. 

        fin: input file name
        keyw: keyword
        """
        f = open(fin, 'r')
        if keyw == '$misc':
            #-- (optional) misc settings
            f.seek(0,0)
            if len(gbutil.gbgrep2(f, keyw, comment='#'))!=0:
                for attr in ['geomformat','freqformat','gradformat']:
                    f.seek(0,0)
                    line = gbutil.gbgrep2(f, attr, comment='#')
                    if len(line) != 0:
                        self.misc[attr] = \
                            re.search(attr+r'=(\S+)', line).group(1)
        elif keyw == '$natom':
            #-- number of atoms
            f.seek(0,0)
            if len(gbutil.gbgrep2(f, keyw, comment='#')) == 0:
                raise ValueError('%s not found. Abort.'%(keyw,))
            self.natom = int(gbutil.skipcomment(f, '#').split()[0])
            self.ncart = self.natom*3
        elif keyw == '$geomgs':
            #-- atomic number and geometry of GS 
            f.seek(0,0)
            if len(gbutil.gbgrep2(f, keyw, comment='#')) == 0:
                raise ValueError('%s not found. Abort.'%(keyw,))
            self.atomnum, x0 = [], []
            if self.misc['geomformat'] == 'g09':
                #-- In 'standard orientation'
                #--   format of g09 output ('dummy atomnum dummy x y z')
                for i in range(self.natom):
                    line = gbutil.skipcomment(f, '#').split()
                    for j in range(3):
                        self.atomnum.append(int(line[1]))
                        x0.append(float(line[3+j]))
            elif self.misc['geomformat'] == 'general':
                #-- In general format ('atomsymbol x y z')
                for i in range(self.natom):
                    line = gbutil.skipcomment(f, '#').split()
                    for j in range(3):
                        self.atomnum.append(sym2anum[line[0].lower()])
                        x0.append(float(line[1+j]))
            else:
                raise ValueError('geomformat not recognized. Abort.')
            #-- convert x0 to bohr
            self.x0 = [x*ang2bohr for x in x0]
            #-- get masses
            self.m = map(lambda x:anum2amass[x], self.atomnum) 
        elif keyw == '$geomes':
            #-- (optional) atomic number and geometry of ES 
            #-- If not provided, ES geometry will be extrapolated from GS
            f.seek(0,0)
            xe = []
            if len(gbutil.gbgrep2(f, keyw, comment='#')) != 0:
                self.bgeomes = True
                if self.misc['geomformat'] == 'g09':
                    #-- In 'standard orientation'
                    #--   format of g09 output ('dummy atomnum dummy x y z')
                    for i in range(self.natom):
                        line = gbutil.skipcomment(f, '#').split()
                        for j in range(3):
                            xe.append(float(line[3+j]))
                elif self.misc['geomformat'] == 'general':
                    #-- In general format ('atomsymbol x y z')
                    for i in range(self.natom):
                        line = gbutil.skipcomment(f, '#').split()
                        for j in range(3):
                            xe.append(float(line[1+j]))
                else:
                    raise ValueError('geomformat not recognized. Abort.')
                #-- convert xe to bohr
                self.xe = [x*ang2bohr for x in xe]
        elif keyw == '$freqes':
            #-- frequencies and normal modes of ES from freq=hpmodes output
            #--   or adf output, in cm-1. 
            f.seek(0,0)
            if len(gbutil.gbgrep2(f, keyw, comment='#')) == 0:
                raise ValueError('%s not found. Abort.'%(keyw,))
            self.freq, eigvec = [], []
            if self.misc['freqformat'] == 'g09':
                for i in range((3*self.natom-6-1)//5+1):
                    line = gbutil.gbgrep2(f, 'Frequencies').split()
                    for j in range(min(3*self.natom-6-5*i, 5)):
                        tmp = float(line[2+j])
                        if tmp>0.0: self.freq.append(tmp)
                        else: 
                            #-- ignore imag freq (which causes problem)
                            # freq.append(1e-10)
                            #-- or flip imag freq
                            self.freq.append(-tmp) 
                    tmp = []
                    gbutil.gbgrep2(f, 'Coord Atom')
                    for k in range(self.ncart):
                        line = f.readline().split()
                        tmp.append(line[3:])
                    tmp = np.transpose(tmp)
                    for j in range(min(3*self.natom-6-5*i, 5)):
                        eigvec.append(tmp[j])
            elif self.misc['freqformat'] == 'adf':
                for i in range((3*self.natom-6-1)//3+1):
                    line = gbutil.skipcomment(f).split()
                    for j in range(min(3*self.natom-6-3*i, 3)):
                        tmp = float(line[j])
                        if tmp>0.0: self.freq.append(tmp)
                        else: 
                            #-- ignore imag freq (which causes problem)
                            # freq.append(1e-10)
                            #-- or flip imag freq
                            self.freq.append(-tmp) 
                    tmp = []
                    f.readline()
                    for k in range(self.natom):
                        line = f.readline().split()
                        tmp.append(line[1:])
                    tmp = np.array(tmp, dtype=float)
                    for j in range(min(3*self.natom-6-3*i, 3)):
                        tmp2 = np.reshape(tmp[:,j*3:(j+1)*3],-1,order='C')
                        eigvec.append(tmp2)

            else: 
                raise ValueError('freqformat not recognized. Abort.')
            eigvec = np.array(eigvec, float)
            #-- eigenvector in mass-weighted Cartesian (and normalized)
            tmp = np.dot(eigvec, np.diag([np.sqrt(x) for x in self.m])) 
            self.eigvec = np.zeros([self.ncart-6, self.ncart], float)
            for i in range(self.ncart-6):
                self.eigvec[i] = tmp[i]/np.linalg.norm(tmp[i])
        elif keyw == '$forcees':
            #-- gradients of ES from g09 output 'Forces (Hartrees/Bohr)' format
            #--   This is needed only when ES geometry is not given
            f.seek(0,0)
            if len(gbutil.gbgrep2(f, keyw, comment='#')) == 0:
                return
            grad = []
            for i in range(self.natom):
                line = gbutil.skipcomment(f, '#').split()
                for j in range(3):
                    grad.append(-float(line[2+j]))
            #-- gradient in mass-weighted Cartesian in hartree/bohr/sqrt(amu)
            self.grad = map(lambda x,y: x/np.sqrt(y), grad, self.m)
        elif keyw == '$unit':
            #-- (optional) unit of the following input energies
            f.seek(0,0)
            if len(gbutil.gbgrep2(f, keyw, comment='#')) != 0:
                self.unit = gbutil.skipcomment(f, '#').strip().lower()
        elif keyw == '$evert':
            #-- vertical excitation
            f.seek(0,0)
            if len(gbutil.gbgrep2(f, keyw, comment='#')) == 0:
                return
            self.Evert = float(gbutil.skipcomment(f, '#'))
        elif keyw == '$oscstr':
            #-- (optional) oscillator strength
            f.seek(0,0)
            if len(gbutil.gbgrep2(f, keyw, comment='#')) == 0:
                self.oscstr = 1.0
                return
            self.oscstr = float(gbutil.skipcomment(f, '#'))
        elif keyw == '$tdm2':
            #-- (optional) transition dipole moment magnitude squared
            f.seek(0,0)
            if len(gbutil.gbgrep2(f, keyw, comment='#')) == 0:
                self.tdm2 = 1.0
                return
            self.tdm2 = float(gbutil.skipcomment(f, '#'))
        elif keyw == '$plot':
            #-- (optional) range of band to plot: (start) (end) (# of points)
            #--   If not given, the band won't be printed
            self.start = self.end = self.npoints = None
            f.seek(0,0)
            if len(gbutil.gbgrep2(f, keyw, comment='#')) != 0:
                line = gbutil.skipcomment(f, '#').split()
                [self.start, self.end] = map(float, line[:2])
                self.npoints = int(line[2])


    def xe_q0(self):
        """Extrapolate from x0 to xe if necessary. 
        Calculate mass-weighted Cart. displacement q0. 
        Calculate deltaE = E_vert - E_adiab. 
        """
        #-- if xe is read from input, 
        #--   then calculate q0 from xe-x0
        #-- otherwise calculate q0 by extrapolation
        #--   and xe from q0 and x0
        if self.bgeomes:
            self.q0 = np.array(map(lambda x,y,z: np.sqrt(x)*(y-z),
                                   self.m, self.xe, self.x0))
        else:
            #-- freq in cm-1 -> omega in s**-1 -> omega**2 in s**-2 = j/m**2/kg
            #-- -> force constant in hartree/bohr**2/amu
            fc = np.array([(x*100*2*np.pi*c)**2/hartree2joule*bohr2meter**2*amu2kg
                           for x in self.freq])
            #-- invert fc and form diagonal matrix
            fcinv = np.diag([x**-1 for x in fc])
            #-- predict equilibrium
            self.q0 = -np.dot(np.transpose(self.eigvec), 
                              np.dot(fcinv, np.dot(self.eigvec, self.grad)))
            self.xe = np.array(map(lambda x,y,z: 1/np.sqrt(x)*y+z,
                                   self.m, self.q0, self.x0))
            #-- energy difference
            self.deltaE = 0.5*np.dot(self.grad, np.dot(np.transpose(self.eigvec),
                                    np.dot(fcinv, np.dot(self.eigvec, self.grad))))
    

    @classmethod
    def from_input(cls, fin):
        """Read all stuff from input.

        fin: string of input file name. 
        """
        inst = cls()
        inst.readstuff(fin, '$natom')
        inst.readstuff(fin, '$misc')
        inst.readstuff(fin, '$geomgs')
        inst.readstuff(fin, '$geomes')
        inst.readstuff(fin, '$freqes')
        inst.readstuff(fin, '$forcees')
        inst.readstuff(fin, '$unit')
        inst.readstuff(fin, '$evert')
        inst.readstuff(fin, '$oscstr')
        inst.readstuff(fin, '$tdm2')
        inst.readstuff(fin, '$plot')
        return inst


    def calc_band_param(self):
        """Calculate parameters needed for simulating band. 
        
        """
        #-- run xe_q0 if not done yet
        if self.q0 is None:
            self.xe_q0()

        #-- displacement of geometry in dimensionless normal coordinates
        #-- work in SI
        #---- q0 in sqrt(amu)*bohr -> sqrt(kg)*meter (SI unit)
        q0SI = self.q0*np.sqrt(amu2kg)*bohr2meter  #np.diag([np.sqrt(x*amu2kg) for x in m])
        #---- omega in cm-1 -> omega**1/2 in (joule/m**2/kg)**1/4
        sqrtomega = np.diag([np.sqrt(x*100*2*np.pi*c) for x in self.freq])
        #---- displacement of geometry in dimensionless normal coordinates
        bigq = np.dot(sqrtomega, np.dot(self.eigvec, q0SI))/np.sqrt(hbar)

        #-- parameters alpha, gamma for vibronic shape model (freq in eV)
        self.bigq = bigq
        self.al = sum(map(lambda x,y:x**2*(y/ev2cm_1)**2, bigq, self.freq))/4
        self.ga = sum(map(lambda x,y:x**2*(y/ev2cm_1)**3, bigq, self.freq))/12



    def plot(self, mode='print', norm='oscstr', 
             start=None, end=None, npoints=None, 
             order=0):
        """Plot spectra or return an array containing the spectral data. 

        mode: 'print'= print the data for plot; 'array'= return the data array. 
        norm: band multiplied by osc. str ('oscstr') or 
              transition dipole moment squared ('tdm') times excitation energy.
        start, end: range of plot. 
        npoints: number of points in the range. 
        order: band shape model type. 
                0 = Gaussian FC-DHO
               -1 = third-order FC-DHO
        return: None if mode='print'; 
                numpy array if mode='array', 
                  with col1 = energy, col2 = abs. strength
        """
        if start is None and self.start is None: 
            raise ValueError('Range of plot not set. Stop.')
        elif start is None:
            start = self.start
            end = self.end
            npoints = self.npoints
        if self.al is None:
            self.calc_band_param()
        al = self.al
        ga = self.ga
        Evert = self.Evert
        unit = self.unit

        if mode == 'print':
            print 'Simulated band'
            print '%7s %10s'%('E', 'abs')
        elif mode == 'array':
            array = []
        for i in range(npoints+1):
            omega = start+(end-start)/npoints*i
            absstr = self.calc_band_point(omega, order, norm)
            if mode == 'print':
                print '%7.4f % 10.4e'%(omega, absstr)
            elif mode == 'array':
                array.append([omega, absstr])
        if mode == 'array':
            return np.array(array)


    def printstuff(self, stuff):
        """Print something. 

        stuff: string of what to print. 
            'geomgs' = ground-state equilibrium geometry (x0)
            'geomes' = excited-state equilibrium geometry (xe)
            'dE' = difference of ES energy at x0 and xe
            'Q' = Dimensionless normal coordinate displacement x0->xe
            'bandparam' = band parameters
            'freq' = harmonic frequencies
            'header' = header
        """
        if stuff == 'geomgs':
            x0byatom = np.reshape(self.x0, (self.natom,3))/ang2bohr
            print '\n==== Ground-State Equilibrium Geometry ===='
            print 'GS geometry, x0:'
            for i in range(self.natom):
                print '% 11.6f % 11.6f % 11.6f'\
                    %(x0byatom[i][0],x0byatom[i][1],x0byatom[i][2])
        elif stuff == 'geomes':
            #-- center of atomic number
            anbyatom = [np.reshape(self.atomnum, (self.natom,3))[i][0]
                        for i in range(self.natom)]
            xebyatom = np.reshape(self.xe, (self.natom,3))
            com = np.zeros([3], float)
            com[0] = sum([anbyatom[i]*xebyatom[i][0]/sum(anbyatom) 
                          for i in range(self.natom)])
            com[1] = sum([anbyatom[i]*xebyatom[i][1]/sum(anbyatom) 
                          for i in range(self.natom)])
            com[2] = sum([anbyatom[i]*xebyatom[i][2]/sum(anbyatom) 
                          for i in range(self.natom)])
            for i in range(self.natom):
                for j in range(3):
                    xebyatom[i][j] = xebyatom[i][j]-com[j]
            xebyatom = xebyatom/ang2bohr
            print '\n==== Excited-State Equilibrium Geometry ===='
            if self.bgeomes:
                print 'ES geometry (read-in), xe:'
            else:
                print 'ES geometry (extrapolated), xe:'
            for i in range(self.natom):
                print '% 11.6f % 11.6f % 11.6f'\
                    %(xebyatom[i][0],xebyatom[i][1],xebyatom[i][2])
        elif stuff == 'dE':
            print '\n==== Difference of ES energy at x0 and xe  ===='
            if not self.bgeomes:
                print 'En(x0)-En(xe) = %11.6f hartree = %11.6f eV'\
                    %(self.deltaE, self.deltaE*hartree2ev)
            else:
                print 'dE undefined.'
        elif stuff == 'Q':
            print '\n==== Dimensionless normal coordinate displacement ===='
            print 'Dimensionless normal coordinate displacement, x0->xe:'
            for i in range(len(self.bigq)):
                print '%3d % 11.6f'%(i+1, self.bigq[i])
        elif stuff == 'bandparam':
            self.calc_band_param()
            #-- std dev of Gaussian in vibronic model in cm-1
            stddev = np.sqrt(sum(map(lambda x,y:x**2*y**2, 
                                     self.bigq, self.freq))/2)
            al = self.al
            ga = self.ga
            print '\n==== Band Parameters ===='
            print '* Gaussian FC-DHO model:'
            print 'Std dev = % 11.6f hartree = % 11.6f eV = % 11.6f cm-1'\
                %(stddev/hartree2cm_1, stddev/hartree2cm_1*hartree2ev, stddev)
            print '   FWHM = % 11.6f hartree = % 11.6f eV = % 11.6f cm-1'\
                %(2.35482*stddev/hartree2cm_1, 2.35482*stddev/hartree2cm_1*hartree2ev, 2.35482*stddev)
            print '* Third-order FC-DHO model:'
            print 'Parameters: alpha = % 11.6f eV**2, gamma = % 11.6f eV**3'%(al, ga)
            print 'Peak shift based on the third-order model: % 11.6f eV'\
                %((2*al**2-np.sqrt(4*al**4+18*al*ga**2))/3/ga)
        elif stuff == 'freq':
            print '\n==== Frequencies ===='
            print '%5s %11s'%('Mode', 'Freq (cm-1)')
            for i in range(len(self.freq)):
                print '%5d %11.6f'%(i+1, self.freq[i])
        elif stuff == 'header':
            print '''
======================================================================
==                             FCBand                               ==
==   Franck-Condon displaced harmonic oscillator (FC-DHO) models    ==
==   for simulation of vibronic bands of absorption spectroscopy    ==
==                                                                  ==
==                          Version 2017                            ==
==                                                                  ==
==                Shaohong L. Li and Donald G. Truhlar              ==
==                      University of Minnesota                     ==
======================================================================
            '''


    def calc_band_point(self, omega, order=0, norm='one'):
        """Compute one point of band at photon energy omega.

        omega: photon energy in unit of self.unit
        order: band shape model type. 
                0 = Gaussian FC-DHO
               -1 = third-order FC-DHO
        norm: normalization/scale.
              ='one': no scale
              ='oscstr': scale by oscillator strength
              ='tdm': scale by transition dipole moment squared times excitation energy
        """
        al = self.al
        ga = self.ga
        Evert = self.Evert
        unit = self.unit
        #-- omega must have the same unit as the E's
        beta = omega-Evert

        #-- convert beta to eV if needed since al and ga are in eV
        if unit == 'cm_1':
            beta /= ev2cm_1
        elif unit != 'ev':
            raise ValueError('Unrecognized unit: %s'%unit)

        #-- third-order FC-DHO
        if order == -1:
            integrand = lambda t, beta: exp(-al*t**2)*cos(beta*t+ga*t**3)
            integral = lambda beta: quad(integrand, 0, Inf, args=(beta))
            res = integral(beta)[0]
        #-- Gaussian FC-DHO
        elif order == 0:
            res = (pi/al)**0.5*exp(-beta**2/4/al)
        else:
            raise ValueError('Unknown order.')
        
        #-- normalization
        if norm == 'oscstr':
            res *= self.oscstr
        elif norm == 'tdm':
            res *= self.tdm2*omega
        elif norm != 'one':
            raise ValueError('norm must be one of "one", "oscstr", "tdm".')
        return res
    
    

