"""Definition of GaussSpectrum class.

"""

import numpy as np
from fcband import asymmband
from fcband.gbconst import *
import copy
from scipy.optimize import minimize

class GaussSpectrum(object):
    """Class of a container of AsymmBand to form full spectrum of a system.

    """
    def __init__(self):
        self.bands = []
        self.exp_spectrum = None

        self.spectrum = None
        self.scaled_spectrum = None

    def add_band(self, band):
        """Add a band to the spectrum.

        band: AsymmBand object
        """
        self.bands.append(band)

    def read_exp_spectrum(self, filename, energy_unit='cm_1', 
                          normalize_to=None):
        """Read in an experimental spectrum from a file.

        The file should have two columns, 
        where col 1 is energy and col 2 is absorption strength

        filename: input file name of experimental spectrum
        energy_unit: energy unit of experimental spectrum;
                           'cm_1' or 'eV' or 'nm'
        normalize_to: whether and how to normalize the exp. spectrum 
                      ='unit_area': normalize to unit area
                      ='unit_maxstr': normalize to unit max abs
                      =None: don't normalize
        """

        #-- exp_spectrum will be an n_points * 2 array
        #--   where col 1 is energy and col 2 is absorption strength
        self.exp_spectrum = []
        for line in open(filename):
            energy, strength = [float(x) for x in line.split()]
            #-- convert to eV (internal energy unit)
            if energy_unit == 'cm_1':
                energy /= ev2cm_1
            elif energy_unit == 'nm':
                energy = nm_times_ev/energy
            self.exp_spectrum.append([energy, strength])
        #-- sort in increasing eV order
        self.exp_spectrum.sort(key=lambda x: x[0])

        #-- normalization
        if normalize_to is not None:
            self.exp_spectrum = GaussSpectrum.normalize(
                self.exp_spectrum, normalize_to)

    def gen_spectrum(self, xs=None, use_exp_xs=False, norm='oscstr',
                     order=0):
        """Generate spectrum.
        
        xs: a list of energy points at which the spectrum is generated.
        use_exp_xs: whether use energy points of the exp. band as xs
        norm: spectrum multiplied by osc. str ('oscstr') or 
              transition dipole moment squared ('tdm') times excitation energy.
        order: band shape model type. 
                0 = Gaussian FC-DHO
               -1 = third-order FC-DHO
        """
        if use_exp_xs:
            if self.exp_spectrum is None:
                raise ValueError('use_exp_xs is True but no exp band exists.')
            xs = [x[0] for x in self.exp_spectrum]
        elif xs is None:
            if not self.bands or self.bands[0].start is None:
                raise ValueError('No xs found.')
            else:
                start = self.bands[0].start
                end = self.bands[0].end
                npoints = self.bands[0].npoints
                xs = np.arange(start, end, float(end-start)/npoints)
        
        self.spectrum = []
        for omega in xs:
            strength = 0
            for band in self.bands:
                strength += band.calc_band_point(omega, order, norm)
            self.spectrum.append([omega, strength])

    def print_spectrum(self, which_spectrum='sim', energy_unit='eV',
                       xs=None, use_exp_xs=False, norm='oscstr',
                       order=0):
        """

        which_spectrum: which to print; 'sim' or 'exp' or 'scaled_sim'
        energy_unit: energy unit in the print; 'eV' or 'cm_1' or 'nm'
        xs: a list of energy points at which the spectrum is generated.
        use_exp_xs: whether use energy points of the exp. band as xs
        norm: spectrum multiplied by osc. str ('oscstr') or 
              transition dipole moment squared ('tdm') times excitation energy.
        order: band shape model type. 
                0 = Gaussian FC-DHO
               -1 = third-order FC-DHO
        """
        if which_spectrum == 'sim':
            if self.spectrum is None:
                self.gen_spectrum(xs, use_exp_xs, norm, order)
            spectrum = self.spectrum
        elif which_spectrum == 'exp':
            spectrum = self.exp_spectrum
        elif which_spectrum == 'scaled_sim':
            spectrum = self.scaled_spectrum
        else:
            raise ValueError('which_spectrum not recognized.')

        print '%8s %12s'%('E (%s)'%energy_unit, 'abs')
        for energy, strength in spectrum:
            if energy_unit == 'cm_1':
                energy *= ev2cm_1
            elif energy_unit == 'nm':
                energy = nm_times_ev/energy
            print '%12.4f %12.6f'%(energy, strength)

    def scale_sim_to_exp(self):
        """Scale simulated spectrum to experimental to minimize the mean squared difference.
        
        return: final mean squared difference
        """
        #-- re-gen simulated spectrum using energy points from exp
        self.gen_spectrum(use_exp_xs=True, norm='oscstr')
        
        def obj_fun(N):
            diff = 0.
            sim = self.spectrum
            exp = self.exp_spectrum
            for i in xrange(len(sim)):
                diff += (N*sim[i][1]-exp[i][1])**2
            return diff/len(sim)

        res = minimize(obj_fun, x0=[1.], method='BFGS', jac=False)
        if not res.success:
            raise RuntimeError('Optimization failed. Message: %s'
                               %res.message)
        N = res.x
        self.scaled_spectrum = [[energy, N*strength] 
                                for energy, strength in self.spectrum]
        return obj_fun(N)

    @staticmethod
    def normalize(spectrum, normalize_to):
        """Normalize spectrum to unit area (regardless of energy unit).

        spectrum: npoints * 2 array; col1 is energy, col2 is strength
        normalize_to: how to normalize
            'unit_area' = normalize to unit area
            'unit_maxstr' = normalize maximum strength to one
        return: npoints * 2 array of normalized spectrum
        """
        res = copy.deepcopy(spectrum)
        if normalize_to == 'unit_area':
            factor = GaussSpectrum.calc_spectrum_area(spectrum)
        elif normalize_to == 'unit_maxstr':
            factor = max([x[1] for x in spectrum])
        else:
            raise ValueError('Normalization not recognized.')

        for i in xrange(len(res)):
            res[i][1] /= factor
        return res

    @staticmethod
    def calc_spectrum_area(spectrum):
        #-- trapezoidal rule
        area = 0.
        for i in xrange(len(spectrum)-1):
            area += (spectrum[i+1][0]-spectrum[i][0])*(spectrum[i+1][1]+spectrum[i][1])
        area /= 2.
        return area
